#include <cstddef>
#include <fstream>
#include <ostream>
#include <string>
#include <vector>

#include "gsl/multi_span"

#include "state.hpp"

class WireWorld {
	std::size_t _noCols{};
	std::size_t _noRows{};
	std::vector<State> _memoryA;
	std::vector<State> _memoryB;
	gsl::multi_span<State, -1, -1> _model;
	gsl::multi_span<State, -1, -1> _nextModel;

  public:
	WireWorld(std::istream& input) {
		using namespace std::string_literals;

		input >> _noCols >> _noRows;

		std::string dummy;
		std::getline(input, dummy);

		if (_noCols < 1 || _noRows < 1)
			throw std::logic_error("File header corrupt");

		_memoryA = std::vector<State>((_noCols + 2) * (_noRows + 2));
		_memoryB = std::vector<State>((_noCols + 2) * (_noRows + 2));
		_model = gsl::as_multi_span(_memoryA.data(), gsl::dim(_noCols + 2),
		                            gsl::dim(_noRows + 2));
		_nextModel = gsl::as_multi_span(_memoryB.data(), gsl::dim(_noCols + 2),
		                                gsl::dim(_noRows + 2));

		for (std::size_t y{1}; y <= _noRows; ++y) {
			std::string row;
			std::getline(input, row);
			if (input.bad())
				throw std::logic_error("Unexpected EOF at line:"s +
				                       std::to_string(y + 1));
			if (row.size() != _noCols)
				throw std::logic_error("Line length not matching at line:"s +
				                       std::to_string(y + 1));

			for (std::size_t x{1}; x <= _noCols; ++x) {
				_model[x][y] = to_enum<State>(row[x - 1]);
			}
		}
	}

	friend std::ostream& operator<<(std::ostream& out, const WireWorld& g) {
		for (std::size_t y{1}; y <= g._noRows; ++y) {
			for (std::size_t x{1}; x <= g._noCols; ++x) {
				out << to_integral(g._model[x][y]);
			}
			out << '\n';
		}
		return out;
	}

	void simulateStep(std::size_t noSteps = 1) {
		for (std::size_t step{0}; step < noSteps; ++step) {
			for (std::size_t y{1}; y <= _noRows; ++y) {
				for (std::size_t x{1}; x <= _noCols; ++x) {
					auto nextState = _model[x][y];
					switch (_model[x][y]) {
					case State::ElectronHead:
						nextState = State::ElectronTail;
						break;
					case State::ElectronTail:
						nextState = State::Conductor;
						break;
					case State::Conductor: {
						const std::array<State, 9> mooreNeighborhood = {
						    _model[x - 1][y - 1], //
						    _model[x + 0][y - 1], //
						    _model[x + 1][y - 1], //
						    _model[x - 1][y + 0], //
						    _model[x + 1][y + 0], //
						    _model[x - 1][y + 1], //
						    _model[x + 0][y + 1], //
						    _model[x + 1][y + 1]  //
						};

						const auto& headCount = std::count(
						    std::begin(mooreNeighborhood),
						    std::end(mooreNeighborhood), State::ElectronHead);

						nextState = (1 == headCount || headCount == 2)
						                ? State::ElectronHead
						                : State::Conductor;

					} break;
					default:
						break;
					}
					_nextModel[x][y] = nextState;
				}
			}
			std::swap(_model, _nextModel);
		}
	}
};
